#pragma once

#include <amdis/common/TypeTraits.hpp>
#include <amdis/linearalgebra/PreconditionerInterface.hpp>

namespace AMDiS
{
  class SolverInfo;

  /// Interface for Runner / Worker types used in solver classes
  template <class Matrix, class VectorX, class VectorB = VectorX>
  class RunnerInterface
  {
  protected:
    using PreconBase = PreconditionerInterface<Matrix, VectorX, VectorB>;

  public:
    /// virtual destructor
    virtual ~RunnerInterface() = default;

    /// Is called at the beginning of a solution procedure
    virtual void init(Matrix const& A) = 0;

    /// Is called at the end of a solution procedure
    virtual void exit() = 0;

    /// Solve the system A*x = b
    virtual int solve(Matrix const& A, VectorX& x, VectorB const& b,
                      SolverInfo& solverInfo) = 0;

    /// Solve the adjoint system A^T*x = b
    virtual int adjointSolve(Matrix const& A, VectorX& x, VectorB const& b,
                             SolverInfo& solverInfo)
    {
      error_exit("Must be implemented by derived class.");
      return 0;
    }

    virtual std::shared_ptr<PreconBase> leftPrecon() { return {}; }
    virtual std::shared_ptr<PreconBase> rightPrecon() { return {}; }
  };

} // end namespace AMDiS
